/*
 * Nachtschaltung.c, Scott-Falk Hhn, v1.00, 21.03.2021
 * Erfasst analog an PB4 die Helligkeit und schaltet ber den Ausgang PB3 das Display und den Touchscreen-Controller fr das Anzeigemodul 4
 *
 * Taktfrequenz 1,0 MHz
 *
 * Belegung der I/O-Ports:
 * PB0 - ISP-MOSI
 * PB1 - ISP-MISO
 * PB2 - ISP-SCK
 * PB3 - Schaltausgang
 * PB4 - Analogeingang
 * PB5 - ISP-RESET
 *
 * SELFPRGEN = [ ]
 * RSTDISBL  = [ ]
 * DWEN      = [ ]
 * SPIEN     = [X]
 * WDTON     = [ ]
 * EESAVE    = [ ]
 * BODLEVEL  = 4V3
 * CKDIV8    = [X]
 * CKOUT     = [ ]
 * SUT_CKSEL = INTRCOSC_8MHZ_6CK_14CK_64MS
 *
 * EXTENDED  = 0xFF
 * HIGH      = 0xDC
 * LOW       = 0x62
 */

#define F_CPU 1000000UL                                                     // Taktfrequenz 1MHz
#include <avr/io.h>
#include <avr/interrupt.h>

#define BRIGHTNUM 16                                                        // Anzahl der Speicherpltze im Helligkeitswerte-Feld
#define BRILEVELH 11 * 256 / 50                                             // Helligkeits-Schwelle fr Einschalten (1,1V)
#define BRILEVELL 9 * 256 / 50                                              // Helligkeits-Schwelle fr Ausschalten (0,9V)
#define TIMESTART 90                                                        // Sperrzeit beim Systemstart
#define TIMEON 30                                                           // Sperrzeit beim Einschalten
#define TIMEOFF 30                                                          // Sperrzeit beim Ausschalten
#define OUTPORT PINB3                                                       // IO-Port fr Schaltausgang

volatile uint16_t mscount;                                                  // Millisekunden-Zhler
uint8_t scount;                                                             // Sekunden-Zhler
uint8_t bright[BRIGHTNUM];                                                  // Feld mit den letzten Helligkeitswerten
uint8_t bripos;                                                             // aktuelle Feldposition
uint8_t i;                                                                  // Zhler-Variable
uint16_t sum;                                                               // Summen-Zwischenspeicher
uint8_t briavg;                                                             // Helligkeits-Mittelwert

ISR(TIMER0_COMPA_vect, ISR_NOBLOCK)                                         // Timer0: Output Compare A Match Interrupt, Routine unterbrechbar, Takt: 1000Hz / 1ms
{                                                                           // steuert Sperrzeit-Zhler und liest Helligkeitswerte vom AD-Wandler
  mscount ++;                                                               // Millisekunden-Zhler erhhen
  if (mscount > 999) {                                                      // wenn Endwert (1000ms) erreicht
    mscount = 0;                                                            // Millisekunden-Zhler zurcksetzen
    if (scount > 0) {                                                       // wenn Sekunden-Zhler aktiv
      scount --;                                                            // Sekunden-Zhler vermindern
    }
  }
  if ((mscount % 10) == 0) {                                                // wenn 10ms-Raster erreicht
    bright[bripos ++] = ADCH;                                               // aktuellen Helligkeitswert in Feld speichern, Feldposition erhhen
    if (bripos >= BRIGHTNUM) {                                              // wenn letzte Feldposition berschritten
      bripos = 0;                                                           // auf erste Feldposition setzen
    }
  }
}

void start_timer(uint8_t sec) {                                             // Funktion zum Starten des Sperrzeit-Zhlers (Parameter: Anzahl der Sekunden)
  cli();                                                                    // Interrupts deaktivieren
  mscount = 0;                                                              // Millisekunden-Zhler lschen
  scount = sec;                                                             // Sekunden-Zhler setzen
  sei();                                                                    // Interrupts wieder aktivieren
}

int main(void)                                                              // Initialisierung
{
  PORTB  = _BV(PINB5) | _BV(PINB3) | _BV(PINB2) | _BV(PINB1) | _BV(PINB0);  // PortB: alle Pins auer B4 auf High bzw. mit Pull-up
  DDRB   = _BV(OUTPORT);                                                    // PortB: Ausgabeport ist Ausgang
  TCCR0A = _BV(WGM01);                                                      // Timer0: Mode 2, CTC
  TCCR0B = _BV(CS01);                                                       // Timer0: Vorteiler 8
  OCR0A  = F_CPU / 8 / 1000 - 1;                                            // Timer0: Zhler-Endwert fr 1000Hz setzen
  TIMSK  = _BV(OCIE0A);                                                     // Timer0: Output Compare A Match Interrupt aktivieren
  ADMUX  = _BV(ADLAR) | _BV(MUX1);                                          // ADC: VCC als Referenz, Ergebnis linksbndig, MUX auf ADC2 (PB4)
  ADCSRA = _BV(ADEN) | _BV(ADSC) | _BV(ADATE) | _BV(ADPS2);                 // ADC: aktivieren, starten, Auto-Modus, Vorteiler 16
  sei();                                                                    // Interrupts aktivieren
  start_timer(TIMESTART);                                                   // Sperrzeit-Zhler auf Systemstartzeit setzen und starten

  while (1) {                                                               // Endlosschleife
    sum = 0;                                                                // Summe fr Mittelwertsberechnung lschen
    for (i = 0; i < BRIGHTNUM; i ++) {                                      // alle Helligkeitswerte
      sum += bright[i];                                                     // summieren
    }
    briavg = sum / BRIGHTNUM;                                               // Mittelwert berechnen
    if ((briavg > BRILEVELH) && (~PORTB & _BV(OUTPORT)) && (scount == 0)) { // wenn Einschaltschwelle berschritten, ausgeschaltet und Sperrzeit-Zhler inaktiv
      PORTB |= _BV(OUTPORT);                                                // Schaltausgang aktivieren
      start_timer(TIMEON);                                                  // Sperrzeit-Zhler auf Einschaltzeit setzen und starten
    }
    if ((briavg < BRILEVELL) && (PORTB & _BV(OUTPORT)) && (scount == 0)) {  // wenn Ausschaltschwelle unterschritten, eingeschaltet und Sperrzeit-Zhler inaktiv
      PORTB &= ~_BV(OUTPORT);                                               // Schaltausgang deaktivieren
      start_timer(TIMEOFF);                                                 // Sperrzeit-Zhler auf Ausschaltzeit setzen und starten
    }
  }
}